/* Fichier "algo_middisp.c"
 * Algorithme dit du "midpoint displacement"
 */

 /* --- Principe de l'algorithme --- */
 /*  voir description de midpoint_disp_iterate */
 
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <time.h>

#include "terrain.h"

#define LN_2 0.693147181f

/* ======================================================
 * midpoint_disp_iterate
 *
 * Description: fonction rcursive qui les etapes de 
 *   l'algorithme : fait la moyenne des hauteurs des 4 sommets du carr
 *   donn par les 4 points abcd (coordonnes (i_a, j_a),
 *   (i_b, j_b) etc...), ajoute une valeur aleatoire et place cette valeur 
 *   au milieu du carr (point m).
 *   Les milieux des cots ont pour hauteur la moyenne des deux extremits
 *   de ce cot plus une valeur aleatoire.
 *   La valeur aleatoire est tiree entre -rmax et + rmax, rmax etant 
 *   diminu  chaque recursivit.
 *  On recommence le processus avec les carrs ainsi cres.
 * =====================================================
 */
void
midpoint_disp_iterate (TER_terrain_t * terrain, float rmax, int compteur,
		       int max_iter,
		       int i_a, int j_a, int i_b, int j_b,
		       int i_c, int j_c, int i_d, int j_d)
{
  int i_e, j_e, i_f, j_f, i_g, j_g, i_h, j_h, i_m, j_m;
  float z_e, z_f, z_g, z_h, z_m;	/* altitudes */
  float r;
  TER_param_middisp_t *param = (TER_param_middisp_t *) terrain->spec_param;
  float h = param->h;

  /* calcul des coordonnes des 4 points e, f, g et h, 
      milieux des cots ad, ab, bc et cd */
  i_e = i_a + (i_d - i_a) / 2;
  j_e = j_a;

  i_f = i_a;
  j_f = j_a + (j_b - j_a) / 2;

  i_g = i_b + (i_c - i_b) / 2;
  j_g = j_b;

  i_h = i_c;
  j_h = j_d + (j_c - j_d) / 2;

  /* centre du carr */
  i_m = i_e;
  j_m = j_f;

#ifdef DEBUG_MIDPOINT
  printf ("A=(%d,%d), B=(%d,%d), C=(%d,%d), D=(%d,%d)\n",
	  i_a, j_a, i_b, j_b, i_c, j_c, i_d, j_d);

  printf ("E=(%d,%d), F=(%d,%d), G=(%d,%d), H=(%d,%d)\n",
	  i_e, j_e, i_f, j_f, i_g, j_g, i_h, j_h);
#endif

  /* calcul des hauteurs des points e, f, g et h : 
      moyenne des deux extrmits + valeur aleatoire */
  r = random_float (-rmax, rmax);
  z_e = (TER_get_heightmap_value (terrain, i_a, j_a)
	 + TER_get_heightmap_value (terrain, i_d, j_d)) / 2 + (r);

  r = random_float (-rmax, rmax);
  z_f = (TER_get_heightmap_value (terrain, i_a, j_a)
	 + TER_get_heightmap_value (terrain, i_b, j_b)) / 2 + (r);

  r = random_float (-rmax, rmax);
  z_g = (TER_get_heightmap_value (terrain, i_b, j_b)
	 + TER_get_heightmap_value (terrain, i_c, j_c)) / 2 + (r);

  r = random_float (-rmax, rmax);
  z_h = (TER_get_heightmap_value (terrain, i_c, j_c)
	 + TER_get_heightmap_value (terrain, i_d, j_d)) / 2 + (r);

  /* hauteur du point m : moyenne des hauteurs des 4 sommets abcd 
      + valeur aleatoire */
  r = random_float (-rmax, rmax);
  z_m = (TER_get_heightmap_value (terrain, i_a, j_a)
	 + TER_get_heightmap_value (terrain, i_b, j_b)
	 + TER_get_heightmap_value (terrain, i_c, j_c)
	 + TER_get_heightmap_value (terrain, i_d, j_d)) / 4 + (r);

  TER_set_heightmap_value (terrain, i_e, j_e, z_e);
  if (z_e > terrain->max_height)
    terrain->max_height = z_e;

  TER_set_heightmap_value (terrain, i_f, j_f, z_f);
  if (z_f > terrain->max_height)
    terrain->max_height = z_f;

  TER_set_heightmap_value (terrain, i_g, j_g, z_g);
  if (z_g > terrain->max_height)
    terrain->max_height = z_g;

  TER_set_heightmap_value (terrain, i_h, j_h, z_h);
  if (z_h > terrain->max_height)
    terrain->max_height = z_h;

  TER_set_heightmap_value (terrain, i_m, j_m, z_m);
  if (z_m > terrain->max_height)
    terrain->max_height = z_m;

  /* diminution de rmax */
  rmax = rmax * (float) (pow (2.0f, -h));
  compteur++;

  /* continue la recursivit avec les sommets des 4 nouveaux carrs :
      afme, fbgm, mgch, emhi */
  if (compteur < max_iter) {

#ifdef DEBUG_MIDPOINT
    printf ("square0\n");
#endif
    midpoint_disp_iterate (terrain, rmax, compteur, max_iter,
			   i_a, j_a, i_f, j_f, i_m, j_m, i_e, j_e);
#ifdef DEBUG_MIDPOINT
    printf ("square1\n");
#endif
    midpoint_disp_iterate (terrain, rmax, compteur, max_iter,
			   i_f, j_f, i_b, j_b, i_g, j_g, i_m, j_m);
#ifdef DEBUG_MIDPOINT
    printf ("square2\n");
#endif
    midpoint_disp_iterate (terrain, rmax, compteur, max_iter,
			   i_m, j_m, i_g, j_g, i_c, j_c, i_h, j_h);

#ifdef DEBUG_MIDPOINT
    printf ("square3\n");
#endif
    midpoint_disp_iterate (terrain, rmax, compteur, max_iter,
			   i_e, j_e, i_m, j_m, i_h, j_h, i_d, j_d);
  }

   /* on a fini, on s'en va */
  else {
#ifdef DEBUG_MIDPOINT
    printf ("j'en ai marre\n");
#endif
    return;
  }
}

void
TER_gen_terrain_middisp (TER_terrain_t * terrain)
{
  int N; /* nombre de fois o l'on appelle la fonction recursive */
  TER_param_middisp_t *param = (TER_param_middisp_t *) terrain->spec_param;
  float m = param->factor / 100.0f;
  float rmax_init = param->rmax_init;
  float nf = 0.0f;
  
  srand (time (NULL));

  /* terrain->size-1 est une puissance de 2, i.e. 2^N
     d'o N = ln (terrain->size-1) / ln(2) */
  nf = (float) (log (terrain->size - 1)) / LN_2;

#ifdef WIN32_BUILD
  /* si on n'a pas la libm */
  N = (int)nf;
#else
  /* plus propre */
  N = (int) rintf (nf);
#endif
  /* on lance la recursivit */
  midpoint_disp_iterate (terrain, rmax_init, 0, N,
			 0, 0, 0, terrain->size - 1,
			 terrain->size - 1, terrain->size - 1,
			 terrain->size - 1, 0);

  TER_FIL_multiply (terrain, m);
  
  printf ("MIDDISP: max_height=%f\n", terrain->max_height);
}
