/* Fichier "algo_particle.c"
 * Algorithme dit de "particle deposition"
 *  (dpot de particules)
 */
 
 /* --- Principe de l'algorithme --- */
 /* simule des pierres qui tombent sur d'autres pierres :
  * si il y a dj quelque chose, la pierre va rouler en bas du tas.
  * On obtient alors un systeme stable.
  */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <time.h>

#include "terrain.h"

#define PARTICLE_UP 0
#define PARTICLE_DOWN 1
#define PARTICLE_RIGHT 2
#define PARTICLE_LEFT 3

#define ALGO_PARTICLE_DONT_ROLL 0
#define ALGO_PARTICLE_ROLL 1

/* ======================================================
 * deposit_particle
 *
 * Description: fonction rcursive qui cherche le bon endroit
 *  o dposer une particule qui tombe initialement en M=(x,y).
 *  La position finale dpend des voisins.
 * =====================================================
 */
 static void
deposit_particle (int x, int y, float dh, TER_terrain_t * terrain)
{

  int j, k, kk = 0, jj = 0;
  int flag = 0;
  float z = 0.0f;

  /* on cherche la position dans les environs : x+-1, y+-1 */
  for (k = -1; k < 2; k++)
    for (j = -1; j < 2; j++)
      /* est-on encore dans le terrain et  une nouvelle place ? */
      if (k != 0 && j != 0 && x + k > -1 && x + k < terrain->size
	  && y + j > -1 && y + j < terrain->size)
      
  /* on cherche  rouler en bas du tas */
	if (TER_get_heightmap_value (terrain, x + k, y + j) <
	    TER_get_heightmap_value (terrain, x, y)) {
        /* un voisin est plus bas, on va rouler. */
        /* il faut continuier les recherches... */
	  flag = 1;
    /* valeurs du deplacement en x et y pour le prochain coup */
	  kk = k; 
	  jj = j;
	}

  if (!flag) {
    /* on a trouv la bonne position */
    z = TER_get_heightmap_value (terrain, x, y) + dh;
    if (z > terrain->max_height)
      terrain->max_height = z;
    TER_set_heightmap_value (terrain, x, y, z);
  }
  else /* on recommence avec le nouveau deplacement kk, jj */
    deposit_particle (x + kk, y + jj, dh, terrain);
}

/* ======================================================
 * TER_gen_terrain_particle
 *
 * Description: boucle sur le nombre d'ilots et sur le 
 *  nombre de particules par ile.
 *  Tire au hasard la position d'une ile, et se deplace tout 
 *  autour en lachant des particules et en les faisant rouler.
 * =====================================================
 */
void
TER_gen_terrain_particle (TER_terrain_t * terrain)
{
  int x, y, dir;
  float z = 0;
  int mode = ALGO_PARTICLE_ROLL;
  TER_param_particle_t *param = (TER_param_particle_t *) terrain->spec_param; 
  int num_iter = param->num_iter;
  int i = 0, num_islands = param->num_islands;
  int iter = 0;
  float dh0 = param->dh0, dh = 0.0f, dhn = 4.2f;
  

  srand (time (NULL));

  for (i = 0; i < num_islands; i++) {

    x = random_int (terrain->size - 1);
    y = random_int (terrain->size - 1);

    for (iter = 0; iter < num_iter; iter++) {

      /* on tire une direction au hasard */
      dir = random_int (4);
      switch (dir) {
      case PARTICLE_UP:
	y--;
	if (y < 0)
	  y = terrain->size - 1;
	break;
      case PARTICLE_DOWN:
	y++;
	if (y > (terrain->size - 1))
	  y = 0;
	break;
      case PARTICLE_LEFT:
	x--;
	if (x < 0)
	  x = terrain->size - 1;
	break;
      case PARTICLE_RIGHT:
	x++;
	if (x > (terrain->size - 1))
	  x = 0;
	break;
      default:
	printf ("Error in particle deposition algorithm\n");
	break;
      }

      /* on depose la particule au bon endroit */
      if (mode == ALGO_PARTICLE_ROLL) {
	deposit_particle (x, y, dh0, terrain);
      }

      else { /* non utilis dans le projet pi4 */
	z = TER_get_heightmap_value (terrain, x, y) + dh0;
	if (z > terrain->max_height)
	  terrain->max_height = z;
	TER_set_heightmap_value (terrain, x, y, z);
      }

      dh = dh0 + (float) ((float) iter / (float) num_iter) * (dhn - dh0);

    }
  }
  TER_FIL_translate_down (terrain);
  printf ("PARTICLE: max_height=%f\n", terrain->max_height);

}
