#include <stdlib.h>
#include <stdio.h>
#include <time.h>

#include "terrain.h"

char
TER_create_new_terrain (TER_terrain_t ** terrain)
{
  *terrain = calloc (1, sizeof (TER_terrain_t));
  if (*terrain == NULL)
    return 0;

  (*terrain)->size = DEFAULT_TERR_SIZE;
  (*terrain)->max_height = 0.0f;

  (*terrain)->heightmap = NULL;
  (*terrain)->heightmap_orig = NULL;
  (*terrain)->spec_param = NULL;
  (*terrain)->generated = 0;
  (*terrain)->algo = DEFAULT_TERR_ALGO;
  (*terrain)->color_scheme = DEFAULT_TERR_COLOR;
  
  srand (time (NULL));

  return 1;
}

char
TER_init_heightmap (TER_terrain_t * terrain)
{
  if (terrain->heightmap) {
    /*fprintf (stderr, "Erasing existing terrain\n");*/
    free (terrain->heightmap);
  }

  terrain->heightmap_orig = (float *) calloc (terrain->size * terrain->size,
					 sizeof (float));
  terrain->heightmap = (float *) calloc (terrain->size * terrain->size,
					 sizeof (float));

  if (terrain->heightmap == NULL || terrain->heightmap_orig == NULL)
    return 0;
  
  if (terrain->spec_param == NULL) {
    fprintf (stderr, "Error: no terrain parameters found.\n");
    return 0;
  }

  TER_reset_filters(terrain);
  
  terrain->max_height = 0.0f;
  terrain->generated = 0;

  return 1;
}

char
TER_gen_terrain (TER_terrain_t * terrain)
{
  if (TER_init_heightmap (terrain)) {

    switch (terrain->algo) {
    case TERR_FLAT:
      break;
    case TERR_MIDDISP:
      TER_gen_terrain_middisp (terrain);
      break;
    case TERR_SPHERE:
      TER_gen_terrain_sphere (terrain);
      break;
    case TERR_HYPERB:
      TER_gen_terrain_hyperbole (terrain);
      break;
    case TERR_DIAM:
      TER_gen_terrain_diamond(terrain);
      break;
    case TERR_FAULT:
      TER_gen_terrain_fault(terrain);
      break;
    case TERR_PARTICLE:
      TER_gen_terrain_particle(terrain);
      break;

    default:
      fprintf (stderr, "Warning: unknown algorithm\n");
      return 0;
      break;
    }
    
    TER_copy_effective_terr_to_orig(terrain);
    
    /* on ne garde que les valeurs au dessus du niveau de la mer */
    TER_FIL_allpositive(terrain);
    
    return 1;
  }

  else {
    fprintf (stderr, "Fatal error occured.\n");
    return 0;
  }
}

char
TER_apply_filter (TER_terrain_t * terrain, int filter, float param)
{
  if (terrain->generated) {

    switch (filter) {
    case TERR_FILTER_NOTHING:
      break;
    case TERR_FILTER_FIR:
      TER_FIL_FIR(terrain, param);
      break;
    case TERR_FILTER_PUISS:
      TER_FIL_power(terrain,param);
      break;
    case TERR_FILTER_MULT:
      TER_FIL_multiply(terrain,param);
      break;
    case TERR_FILTER_CRAT:
      TER_FIL_craters (terrain, param);
      break;
    default:
      fprintf (stderr, "Warning: unknown filter\n");
      return 0;
      break;
    }
    return 1;
  }

  else
    return 0;
}
